# Asterisk::LDAP
# (C) 2005 Ben Klang <ben@alkaloid.net>
# Released under the terms of the GPL.
# See included LICENSE file for more information.
# $Id LDAP.pm,v 0.6.0.1 2005/12/02 13:51:22 bklang Exp $
# Generate configuration files compatible with Asterisk 1.0 from LDAP
#
# Released under the terms of the GPL by Ben Klang (ben@alkaloid.net) 2005-02-12
#
# NOTE NOTE NOTE
# If you get a message similar to
# 'Can't call method "code" without a package or object reference'
# check your TLS settings.  There doesn't appear to be a clean way to catch
# whether start_tls failed, so this program will simply crash rather than send
# unsecured information.
package Asterisk::LDAP;
require 5.006;
#use Exporter;
use File::Temp;
use File::Basename;
use Getopt::Long;
use Net::LDAP;

use strict;
use warnings;

our @ISA = ( 'Asterisk' );

our $VERSION = "0.6.0";
sub version { $VERSION };

=head1 NAME

Asterisk::LDAP - Methods for generating Asterisk configuration from LDAP

=head1 SYNOPSIS

use Asterisk::LDAP;

$AstLDAP = new Asterisk::LDAP;

# parse command line options

        $AstLDAP->parseCmdline();

# parse /etc/ldap.conf (or ldap.conf supplied on command line)

        $AstLDAP->parseLDAPconf();

# parse /etc/asterisk/asterisk.conf to determine best location for generated
# configuration files

        $AstLDAP->parseAsteriskconf();

# Generate the configuration files to temporary files
        foreach $config (@{$AstLDAP->{CONFIGS}}) {
            $tmpfile = $AstLDAP->generate($config, "file");
            rename($tmpfile, $AstLDAP->{CONFIGS}{$config});
        }

# Try to reload Asterisk with the new configuration data

        $AstLDAP->tryreload();

=head1 DESCRIPTION

This module should make it simple to write scripts that customize Asterisk's
configuration based on data from the LDAP tree.  These methods and mechanisms
have been written with customization of the final product (configuration files)
in mind.

=head1 LDAP COMMANDS

=cut

# constructor
sub new {
    my $proto = shift(@_);
    my $class = ref($proto) || $proto;
    my $self = {};
    # These are the only two configurable options that don't really make sense on
    # the command line.  defdialtimeout is the dial timeout to use if LDAP
    # doesn't return a value.  dialprefix is the channel driver and options
    # to prepend to and Dial() string in the dialplan.
    # FIXME DOcument how to set these in globals context in the dialplan
    $self->{'DEFDIALTIMEOUT'} = 90;
    $self->{'DIALPREFIX'} = 'Zap/g9/';
    # End configurable options

    # Begin variable declaration
    @{$self->{'LDAPHOSTS'}} = ();
    $self->{'LDAPPORT'} = undef;
    $self->{'LDAPVERSION'} = undef;
    $self->{'BINDDN'} = undef;
    $self->{'BINDPW'} = undef;
    $self->{'ROOTBINDDN'} = undef;
    $self->{'ROOTBINDPW'} = undef;
    $self->{'BASEDN'} = undef;
    $self->{'SCOPE'} = undef;
    $self->{'OTHER'} = undef;
    $self->{'CACERTDIR'} = undef;
    $self->{'CACERTFILE'} = undef;
    $self->{'ASTETCDIR'} = "/etc/asterisk";
    %{$self->{'CONFIGDATA'}} = (); # Hold read configuration data
    $self->{'CONFIGOVERRIDE'} = 0;
    $self->{'CMDLINE'} = undef;
    $self->{'CMDLINELDAPHOSTS'} = 0;
    $self->{'CMDLINELDAPPORT'} = 0;
    $self->{'CMDLINELDAPVERSION'} = 0;
    $self->{'CMDLINEBINDDN'} = 0;
    $self->{'CMDLINEBINDPW'} = 0;
    $self->{'CMDLINEBASEDN'} = 0;
    $self->{'CMDLINELDAPSSL'} = 0;
    $self->{'CMDLINELDAPTLS'} = 0;
    $self->{'CMDLINECACERTFILE'} = 0;
    $self->{'CMDLINECACERTDIR'} = 0;
    $self->{'CMDLINECHECKPEER'} = 0;
    $self->{'CMDLINEASTETCDIR'} = 0;
    $self->{'LDAPS'} = 0;
    $self->{'STARTTLS'} = 0;
    $self->{'SSLVERIFY'} = 0;
    $self->{'LDAPCONF'} = "/etc/ldap.conf";
    $self->{'ASTERISKCONF'} = "/etc/asterisk/asterisk.conf";
    $self->{'ASTRELOAD'} = 0;
    $self->{'ASTERISKBIN'} = "/usr/sbin/asterisk";
    $self->{'SAVEDCOPIES'} = 3;
    $self->{'UNDO'} = 0;
    $self->{'NEEDRELOAD'} = 0;
    $self->{'PRECOMPILE'} = 0;
    $self->{'FLUSH'} = 0;
    $self->{'CONTEXT'} = undef;
    $self->{'VOICEMAILBOX'} = undef;
    $self->{'NEWPIN'} = undef;
    $self->{'USERBASEDN'} = undef;
    $self->{'GROUPBASEDN'} = undef;
    $self->{'OUTPUT'} = "";
    $self->{'LDAP'} = 0; # Cache the LDAP connection
    # When adding new support for a configuration file, make sure to add its
    # default key and filename to this array.
    @{$self->{'CONFIGS'}} = (
        "extensions",
        "voicemail",
        "musiconhold",
    );

    if (!defined($main::VERBOSE)) {
        $main::VERBOSE = 0;
    }

    # Output handlers
    bless($self);
    return $self;
}

sub debug {
    if (!defined(&main::debug)) {
        my $level = shift;
        my $message = shift;
        if ($main::VERBOSE >= $level) {
            CORE::print STDERR "BUILTIN_DEBUG$level: $message\n";
        }
    } else {
        main::debug(@_);
    }
}

sub print
{
    if (!defined(&main::print)) {
        my $message = shift;
        CORE::print STDERR "$message\n";
    } else {
        main::print(@_);
    }
}

sub warn
{
    if (!defined(&main::warn)) {
        my $message = shift;
        CORE::warn "$message\n";
    } else {
        main::warn(@_);
    }
}

sub die
{
    if (!defined(&main::die)) {
        my $message = shift;
        CORE::die "$message\n";
    } else {
        main::die(@_);
    }
}

=over 4

=item $AstLDAP->getLDAPconf()

Reads and parses into object variables LDAP connection options from ldap.conf

This command will read in the ldap.conf stored in $AstLDAP->ldapconf (defaults
to /etc/ldap.conf) and parse the connection options into variables which can
be used for a subsequent call to $AstLDAP->ldap_connect();

Example: $AstLDAP->ldapconf("/usr/local/etc/ldap.conf"); // set ldap.conf path
         $AstLDAP->getLDAPconf();

Returns: 1 on success.

=cut
sub getLDAPconf
{
    my $self = shift(@_);
    debug(1, "Processing system LDAP configuration");

    # Check for, and possibly use, a root bind password
    my $ldapsecret = $self->{'LDAPCONF'};
    $ldapsecret =~ s/\.conf$/\.secret/;
    if (open(LDAPSECRET, $ldapsecret)) {
        while(<LDAPSECRET>) {
            chomp;
            debug(2, "Setting ROOTBINDPW from $ldapsecret");
            $self->rootbindpw($_);
        }
        close(LDAPSECRET);
    } else {
        debug(2, "No ldap.secret file found.  Will not set ROOTBINDPW");
    }

    # Pull in the host's ldap.conf configuration
    open(LDAPCONF, $self->{LDAPCONF}) or
        &die("Unable to open $self->{LDAPCONF}", __FILE__, __LINE__);
    while (<LDAPCONF>) {
        SWITCH: for ($_) {
        chomp;
        /^host\s+/i && do {
            $_ =~ s/^host\s+//i;
            if ($self->{'CMDLINELDAPHOSTS'} == 0) {
                $self->ldaphosts(split / +/);
            }
            last SWITCH;
        };
        /^port\s+/i && do {
                $_ =~ s/^port\s+//i;
            if ($self->{'CMDLINELDAPPORT'} == 0) { $self->ldapport($_); }
            last SWITCH;
        };
        /^ldap_version\s+/ && do {
            $_ =~ s/^ldapversion\s+//i;
            if ($self->{'CMDLINELDAPVERSION'} == 0) { $self->ldapversion($_); }
            last SWITCH;
        };
        /^binddn\s+/i && do {
            $_ =~ s/^binddn\s+//i;
            if ($self->{'CMDLINEBINDDN'} == 0) { $self->binddn($_); }
            last SWITCH;
        };
        /^bindpw\s+/i && do {
            $_ =~ s/^bindpw\s+//i;
            if ($self->{'CMDLINEBINDPW'} == 0) { $self->bindpw($_); }
            last SWITCH;
        };
        /^base\s+/i && do {
            $_ =~ s/^base\s+//i;
            if ($self->{'CMDLINEBASEDN'} == 0) { $self->basedn($_); }
            last SWITCH;
        };
        /^scope\s+/i && do {
            $_ =~ s/^scope\s+//i;
            if ($self->{'CMDLINESCOPE'} == 0) { $self->scope($_); }
            last SWITCH;
        };
        /^ssl\s+/i && do {
            if ($_ =~ /yes|on|true/i) {
                if ($self->{'CMDLINESSL'} == 0) { $self->ldaps(1); }
            } elsif ($_ =~ /starttls/i) {
                if ($self->{'CMDLINETLS'} == 0) { $self->starttls(1); }
            }
            last SWITCH;
        };
        /^tls_cacertfile\s+/i && do {
            $_ =~ s/^tls_cacertfile\s+//i;
            if ($self->{'CMDLINECACERTFILE'} == 0) { $self->cacertfile($_); }
            last SWITCH;
        };
        /^tls_cacertdir\s+/i && do {
            $_ =~ s/^tls_cacertdir\s+//i;
            if ($self->{'CMDLINECACERTDIR'} == 0) { $self->cacertdir($_); }
            last SWITCH;
        };
        /^tls_checkpeer\s+/i && do {
            if ($_ =~ /yes|on|true/i) {
                if ($self->{'CMDLINECHECKPEER'} == 0) { $self->sslverify(1); }
            } else {
                if ($self->{'CMDLINECHECKPEER'} == 0) { $self->sslverify(0); }
            }
        };
        /^nss_base_passwd\s+/i && do {
            $_ =~ s/^nss_base_passwd\s+//i;
            my @user_formula = split /\?/;
            my $userbasedn = $user_formula[0];
            my $userscope = $user_formula[1];
            my $userfilter = $user_formula[2];
            #FIXME Add support for scope and filter
            $self->userbasedn($userbasedn);
            last SWITCH;
        };
        /^nss_base_group\s+/i && do {
            $_ =~ s/^nss_base_group\s+//i;
            my @group_formula = split /\?/;
            my $groupbasedn = $group_formula[0];
            my $groupscope = $group_formula[1];
            my $groupfilter = $group_formula[2];
            #FIXME Add support for scope and filter
            $self->groupbasedn($groupbasedn);
            last SWITCH;
        };
        /^rootbinddn\s+/i && do {
            $_ =~ s/^rootbinddn\s+//i;
            $self->rootbinddn($_);
            last SWITCH;

        };
        }
    }
    close(LDAPCONF);
}

=item $AstLDAP->getAsteriskconf()

Reads and parses into object variables Asterisk installation information
from asterisk.conf

This command will read in the asterisk.conf stored in $AstLDAP->asteriskconf
(defaults to /etc/asterisk/asterisk.conf) and parse the installation options
into variables which can be used fordetermining where to put the generated
Asterisk configuration files.

Example: // set asterisk.conf path
         $AstLDAP->asteriskconf("/usr/local/etc/asterisk.conf");
         // read and parse the file
         $AstLDAP->getAsteriskconf();

Returns: 1 on success.

=cut
sub getAsteriskconf
{
    my $self = shift(@_);
    if (!$self->{'CMDLINEASTETCDIR'}) {
        open(ASTERISKCONF, $self->{'ASTERISKCONF'}) or
            &die("Unable to open $self->{ASTERISKCONF}", __FILE__, __LINE__);
        while (<ASTERISKCONF>) {
            SWITCH: for ($_) {
                chomp;
                /^astetcdir/ && do {
                    $_ =~ s/astetcdir(\s)*=>(\s)*//;
                    $self->astetcdir($_);
                    last SWITCH;
                };

            }
        }
        close(ASTERISKCONF);
    }
}

=item $AstLDAP->usage()

Print usage information

This function is only useful if using the standard Asterisk::LDAP
parseCmdLine() routines.  Otherwise you will need to supply your own
usage statement.

Example: $AstLDAP->usage();

Returns: die()s printing usage information

=cut
sub usage
{
    my $self = shift(@_);
    &print("\nUsage: $0 [options]\n".
    '
(Note that ALL of these are optional and "intelligent" defaults will be used)

    --host      -h  Set the LDAP host to read from (can be used multiple times)
    --port      -p  Set the LDAP port
    --version   -P  Set LDAP protocol version
    --binddn    -D  Set LDAP bind DN credentials
    --bindpw    -W  Set LDAP bind password credentials
    --basedn    -b  Set LDAP base DN for searches
    --scope     -s  Set LDAP search scope
    -Z[Z]           Request [or require] LDAP TLS security
    --config    -c  Specify an asterisk configuration file to generate
    --reload    -r  Allows automatic \'reload\' of Asterisk after
                    configuration change
    --ldapconf      Specify a path to ldap.conf
    --asterisk      Set the path to the Asterisk binary
    --undo      -u  Rollback last configuration change and exit
    --saved     -i  Set the number of backup copies to be made for rollbacks
    --cacertdir     Sepecify a directory which stores root CA certificates
    --cacertfile    Specify a root CA certificate file to use for TLS/SSL
    --astetcdir     Specify location for Asterisk configuration files
    --quiet     -q  Suppress all output other than errors
    --context       Specify the context to be used
                    (ex. for voicemail pin updates)
    --voicemailbox  Specify voicemail box to be updated
                    (ex. for voicemail pin updates)
    --newpin        Specify new PIN number (ex. for voicemail pin updates)
    --userbasedn    Base DN at which user data is stored
                    (defaults to ou=People,<basedn>)
    --groupbasedn   Base DN at which group data is stored
                    (defaults to ou=Groups,<basedn>)
    --rootbinddn    Bind DN credentials to use when performing administrative
                    actions
    --rootbindpw    Bind password credentials to use when performing
                    administrative actions
    --help          This usage screen
    ');
    exit 1;
}

=item $AstLDAP->parseCmdline()

Configures the Asterisk::LDAP object based on command line options

See $AstLDAP->usage() for parameter documentation

Example: $AstLDAP->parseCmdline();

Returns: 1 on success.

=cut
sub parseCmdline
{
    my $self = shift(@_);
    my $cmdline = new Getopt::Long::Parser;
    $cmdline->configure("bundling");
    $cmdline->getoptions(
                "help" => sub { $self->usage(); },
                "verbose|v+" => sub { $self->verbose(@_); },
                "debug|d+" => sub { $self->verbose(@_); },
                "host|h|H=s@" => sub { $self->ldaphosts(@_) },
                "port|p=i" => sub { $self->ldapport(@_); },
                "version|P=i" => sub { $self->ldapversion(@_); },
                "binddn|D=s" => sub { $self->binddn(@_); },
                "bindpw|W=s" => sub { $self->bindpw(@_); },
                "basedn|b=s" => sub { $self->basedn(@_); },
                "scope|s=s" => sub { $self->scope(@_); },
                "Z+" => sub { $self->starttls(@_); },
                "config|c=s" => sub { $self->configs(@_); },
                "nocolors|n+" => sub { $self->nocolors(@_); },
                "reload|r=i" => sub { $self->astreload(@_); },
                "asterisk=s" => sub { $self->asteriskbin(@_); },
                "undo|u" => sub { $self->needundo(@_); },
                "saved=i" => sub { $self->savedcopies(@_); },
                "cacertdir=s" => sub { $self->cacertdir(@_); },
                "astetcdir=s" => sub { $self->astetcdir(@_); },
                "quiet|q+" => sub { $self->quiet(@_); },
                "context=s" => sub { $self->context(@_); },
                "voicemailbox=s" => sub { $self->voicemailbox(@_); },
                "newpin=s" => sub { $self->newpin(@_); },
                "userbasedn=s" => sub { $self->userbasedn(@_); },
                "groupbasedn=s" => sub { $self->groupbasedn(@_); },
                "rootbinddn=s" => sub { $self->rootbinddn(@_); },
                "rootbindpw=s" => sub { $self->rootbindpw(@_); },
                "ldapconf=s" => sub { $self->ldapconf(@_); },
                "precompile|R+" => sub { $self->precompile(@_); },
                "flush|F+" => sub { $self->flush(@_); },
        );

#foreach my $xhost (@{$self->{CMDLINELDAPHOSTS}}) { print "X$xhost\n"; }
#    if ($self->{CMDLINELDAPHOSTS}) { @{$self->{LDAPHOSTS}} = @{$self->{CMDLINELDAPHOSTS}}; }
}

=item $AstLDAP->readConfig($config)

Reads from Asterisk configuration information from LDAP and populates
the object's internal hash structures for generation.

$config is currently one of
    "extensions"    - For extensions.conf
    "voicemail"     - For voicemail.conf
    "musiconhold"   - For musiconhold.conf

Examples:
            $AstLDAP->readConfig("extensions");

            my @configs = ("extensions", "voicemail", "musiconhold");
            $AstLDAP->readConfig(@configs);

Returns: 1 on success.

=cut
sub readConfig
{
    my $self = shift(@_);
    my $config = shift;
#     my ($line, $entry, $context, %opts, $ou, $extension, %voicehash);
#     my ($user, $users, $number, $mesg);
    my $mesg; # LDAP operations
    my $i = 0;

    if (!defined($self->{'LDAPHOSTS'}[0]) || !defined($self->{'BASEDN'})) {
        &warn("$0: You must specify at least one LDAP host (-h)\n");
        &warn("$0: and the base DN (-b).\n");
        usage();
        exit(1);
    }
    my $ldap;
    $ldap = $self->ldap_connect();

    CONFIG: for ($config) {
    /extensions/ && do {
        my %extensions;
        debug(1, "Reading Asterisk extensions");

        $mesg = $ldap->search(
                base => 'ou=Asterisk,'.$self->{BASEDN},
                filter => "(objectClass=asteriskExtensions)"
            );
        $mesg->code && &die($mesg->error, __FILE__, __LINE__);

        CONTEXT: foreach my $entry ($mesg->entries) {
            debug(3, "Processing ".$entry->dn);
            my $context = $entry->get_value('context');
            $extensions{$context}{'serial'} = $entry->get_value('asteriskSerial');;

            debug(4, "Reading include lines for $context");
            # TODO: Handle the precompile option here
            $i = 0;
            foreach my $line ($entry->get_value('asteriskIncludeLine', 'asref')) {
                $extensions{$context}{'includes'}[$i] = $line;
            }

            debug(4, "Reading exten lines for $context");
            foreach my $line ($entry->get_value('asteriskExtensionLine', 'asref')) {
                my ($exten, $prio, $application) = split(/,/, $line, 4);
                my $args = '';
                my $parenPos = index($application, '(');
                if ($parenPos != -1) {
                    $args = substr($application, $parenPos);
                    $application = substr($application, 0, $parenPos);
                }
                $extensions{$context}{'extensions'}{$exten}{$prio}{'application'} = $application;
                $extensions{$context}{'extensions'}{$exten}{$prio}{'args'} = $args;
            }

            debug(4, "Reading bare lines for $context");
            $i = 0;
            foreach my $line ($entry->get_value("asteriskExtensionBareLine", 'asref')) {
                $extensions{$context}{'barelines'}[$i] = $line;
            }

            debug(4, "Reading ignorepat lines for $context");
            $i = 0;
            foreach my $line ($entry->get_value("asteriskIgnorePat", 'asref')) {
                $extensions{$context}{'ignorepats'}[$i] = $line;
            }
        }
        %{$self->{'CONFIGDATA'}{'extensions'}} = %extensions;
        last CONFIG;
    };

    /voicemail/ && do {
        debug(1, 'Generating voicemail.conf');
        my %timestamps; # Keep track of most recent update user per context
        my $voicemail = \%{$self->{'CONFIGDATA'}{'voicemail'}};

        debug(2, 'Retrieving voicemail contexts');
        $mesg = $ldap->search(
                'base' => $self->{'BASEDN'},
                'filter' => '(objectClass=asteriskVoicemail)',
                'attrs' => [
                    'context',
                    'modifyTimestamp',
                    'asteriskVoicemailBareLine',
                ]
            );
        $mesg->code && &die($mesg->error, __FILE__, __LINE__ );
        foreach my $entry ($mesg->entries) {
            if ($entry->exists('asteriskVoicemailBareLine')) {
                my $context = ($entry->get_value('context'))[0];

                $timestamps{$context} = ($entry->get_value('modifyTimestamp'))[0];
                $timestamps{$context} =~ s/Z$//;
                if (!defined($timestamps{$context}) || !$timestamps{$context}) {
                    &debug(1, 'No timestamp found for '.$context.
                        ', defaulting to 0');
                    $timestamps{$context} = 0;
                } else {
                    &debug(2, "Found timestamp for $context:".
                        $timestamps{$context});
                }

                &debug(3, "Processing bare lines for $context");
                my $counter = 0;
                foreach my $attribute
                    ($entry->get_value('asteriskVoicemailBareLine', 'asref')) {
                    debug(4, "$context: $attribute");
                    $voicemail->{$context}{'barelines'}[$counter] = $attribute;
                    $counter++;
                }
            }
        }
        debug(2, 'Retrieving voicemail users');
        $mesg = $ldap->search(
                'base' => $self->{'BASEDN'},
                'filter' => '(objectClass=asteriskVoiceMailbox)',
                'attrs' => [
                    'context',
                    'voiceMailbox',
                    'modifyTimestamp',
                    'voiceMailboxPin',
                    'cn',
                    'mail',
                    'asteriskPager',
                    'asteriskVoiceMailboxOptions',
                ]
            );
        $mesg->code && &die($mesg->error, __FILE__, __LINE__);

        # Populate a hash with all the voicemail users.  This is
        # necessary because we must associate voicemail boxes with
        # customer contexts and this feels faster than doing multiple
        # LDAP searches
        foreach my $entry ($mesg->entries) {
            debug(3, "Retrieving ".$entry->dn);
            my $context = $entry->get_value('context');
            my $mailbox = $entry->get_value('voiceMailbox');
            my $timestamp = ($entry->get_value('modifyTimestamp'))[0];
            $timestamp =~ s/Z$//;
            &debug(4, "Timestamp for ${mailbox}\@${context}: $timestamp");

            # Check to see if the user is the most recently modified
            # If so use this timestamp as the "serial" number for the context
            if (!defined($timestamps{$context})) {
                &debug(3, "No previous timestamp found for $context.".
                    " Using $timestamp from $mailbox");
                $timestamps{$context} = $timestamp;
            } elsif (defined($timestamp) && $timestamp > $timestamps{$context}) {
                debug(3, "Found newer timestamp for $context: ".
                    $timestamp.' > '.$timestamps{$context});
                $timestamps{$context} = $timestamp;
            } elsif(!defined($timestamp)) {
                &warn("Unable to locate timestamp for $mailbox");
            }

            if (!defined($voicemail->{$context})) {
                &debug(5, "Defining new hash for $context");
                %{$voicemail->{$context}} = ();
            }
            if (!defined($voicemail->{$context}{'mailboxes'})) {
                &debug(5, "Defining new hash for $context/mailbox");
                %{$voicemail->{$context}{'mailboxes'}} = ();
            }
            %{$voicemail->{$context}{'mailboxes'}{$mailbox}} = ();
            my $mdata = \%{$voicemail->{$context}{'mailboxes'}{$mailbox}};

            $mdata->{'pin'} = $entry->get_value('voiceMailboxPin');
            $mdata->{'name'} = $entry->get_value('cn');

            if ($entry->exists('mail')) {
                $mdata->{'email'} = $entry->get_value('mail');
            } else {
                $mdata->{'email'} = '';
            }
            if ($entry->exists('asteriskPager')) {
                $mdata->{'pager'} = $entry->get_value('asteriskPager');
            } else {
                $mdata->{'pager'} = '';
            }

            if ($entry->exists('asteriskVoiceMailboxOptions')) {
                my $counter = 0;
                foreach my $option ($entry->get_value('asteriskVoiceMailboxOptions', 'asref')) {
                    $mdata->{'options'}[$counter] = $option;
                    $counter++;
                }
            } else {
                $mdata->{'options'}[0] = '';
            }
        } # End of foreach user

        foreach my $context (keys(%timestamps)) {
            &debug(4, 'Saving serial number '.$timestamps{$context}.
                ' to '.$context);
            $voicemail->{$context}{'serial'} = $timestamps{$context};
        }

        last CONFIG;
    };
    /musiconhold/ && do {
        debug(1, "Generating musiconhold.conf");
        my $musiconhold = \%{$self->{'CONFIGDATA'}{'musiconhold'}};
        # Initialize to zero as this will hold the most recent update below
        $musiconhold->{'serial'} = 0;
        $mesg = $ldap->search(
                base => $self->{'BASEDN'},
                filter => "(objectClass=asteriskMusicOnHold)"
            );
        $mesg->code && &die($mesg->error, __FILE__, __LINE__ );

        $i = 0;
        foreach my $entry ($mesg->entries) {
            my $context = $entry->get_value("context");
            my $musicclass = $entry->get_value("asteriskMusicOnHoldClass");
            my $serial = $entry->get_value("asteriskSerial");
            if (defined($serial) && $serial > $musiconhold->{'serial'}) {
                &debug(3, "Found newer serial for musiconhold: $serial");
                $musiconhold->{'serial'} = $serial;
            }
            $musiconhold->{'contexts'}{$context} = $musicclass;

        }
        last CONFIG;
    };
    # default:
    &warn("Unknown configuration module: $_");
    } # End of CONFIG block

    # Disconnect from LDAP
    #$mesg = $ldap->unbind;
    #$mesg->code && &warn($mesg->error());
}

=item $AstLDAP->format($mode)

Reads from LDAP and exports Asterisk compatible configuration files.

This command will search through the LDAP directory tree and create Asterisk 1.0
compatible configuration files.

$config is currently one of
    "extensions"    - For extensions.conf
    "voicemail"     - For voicemail.conf
    "musiconhold"   - For musiconhold.conf

$mode can be
    "file"          - Will return a scalar containing the path to a temporary
                      file holding the generated configuration file.
    "scalar"        - Will return a hash of scalars containing the generated file.
                      One hash key for each configuration file read.

Examples:
            $file = $AstLDAP->generate("extensions", "file");

            %configdata = $AstLDAP->generate("scalar");
            # $configdata{'extensions'} holds extensions.conf
            # $configdata{'voicemail'} holds voicemail.conf
            # $configdata{'musiconhold'} hosts musiconhold.conf

Returns: 1 on success.

=cut
sub format
{
    my $self = shift(@_);
    my $config = shift;
    my $mode = shift;
    my ($line, $entry, $context, %opts, $ou, $extension);
    my ($user, $users, $number, $mesg);

    CONFIG: for ($config) {
    /extensions/ && do {
        my %oldSerials;
        debug(2, "Formatting extensions.conf");

        # If we aren't flushing the old config get
        # a list of serials for existing contexts
        if (!$self->{'FLUSH'}) {
            %oldSerials = $self->_getOldSerials('extensions');
        }

        my ($efh, $efile) = $self->_getOutputHandle('file', 'extensions');

        # Process each context and if new write updated data
        CONTEXT: foreach my $context
            (keys(%{$self->{'CONFIGDATA'}{'extensions'}})) {
            my ($fh, $file);
            &debug(3, "Processing context \"$context\"");


            my $cdata = $self->{'CONFIGDATA'}{'extensions'}{$context};
            if (!defined($cdata->{'serial'})) {
                &warn('Serial number not defined for '.$context.': using 0');
                $cdata->{'serial'} = 0;
            }

            # Check to see if we should update the context
            if (defined($oldSerials{$context})) {
                # Remove this context from the list of
                # pre-existing contexts
                if (!$self->{'FLUSH'}) {
                    if ($oldSerials{$context} < $cdata->{'serial'}) {
                        debug(1, "LDAP serial for $context".
                        " ($cdata->{'serial'}) \nis newer than file".
                        " ($oldSerials{$context})");
                    } else {
                        debug(3, "LDAP serial for $context".
                        " ($cdata->{'serial'}) \n is not newer than file".
                        " ($oldSerials{$context}).  Context will not".
                        " be updated.");
                        # Delete the old serial from the set so it
                        # isn't cleaned up later
                        delete $oldSerials{$context};
                        next CONTEXT;
                    }
                }
            } else {
                &debug(2, 'No old serial number found for '.$context.
                    'Assuming new context.');
            }
            # Delete the old serial from the set so it
            # isn't cleaned up later
            delete $oldSerials{$context};

            ($fh, $file) =
                $self->_getOutputHandle('file', 'extensions', $context);

            &debug(2, "Printing context data for $context");
            # Include the context into the master extensions.conf
            CORE::print $efh '#include '.$self->{'ASTETCDIR'}.
                "/extensions.d/extensions.${context}.conf\n";

            CORE::print $fh "[$context] ; $cdata->{'serial'}\n";

            &debug(3, 'Printing extension lines');
            foreach my $exten (keys(%{$cdata->{'extensions'}})) {
                foreach my $prio (sort keys(%{$cdata->{'extensions'}{$exten}})) {
                    my $application = $cdata->{'extensions'}{$exten}{$prio}{'application'};
                    my $args = $cdata->{'extensions'}{$exten}{$prio}{'args'};
                    if (!defined($args)) {
                        # args may be empty
                        $args = '';
                    }
                    &debug(4, "exten => ${exten},${prio},${application}${args}");
                    CORE::print $fh 'exten => '.
                        "${exten},${prio},${application}${args}\n";
                }
            }

            &debug(3, 'Printing include lines');
            foreach my $line (@{$cdata->{'includes'}}) {
                &debug(4, "include => $line");
                CORE::print $fh "include => $line\n";
            }

            &debug(3, 'Printing ignorepat lines');
            foreach my $line
                (@{$cdata->{'ignorepats'}}) {
                &debug(4, "ignorepat => $line");
                CORE::print $fh "ignorepat => $line\n";
            }

            &debug(3, 'Printing bare lines');
            foreach my $line (@{$cdata->{'barelines'}}) {
                &debug(4, "$line");
                CORE::print $fh "$line\n";
            }

            &debug(1, "Wrote all lines for $context, closing and renaming file");
            close($fh);
            rename($file, $self->{'ASTETCDIR'}.'/extensions.d/extensions.'.
                $context.'.conf');
            # We've updated a context.  Flag for reload.
            &debug(1, "Flagging Asterisk for a reload");
            $self->{'NEEDRELOAD'}++;
        } # End of CONTEXT block
        if (!$self->{'NEEDRELOAD'}) {
            &debug(2, "No changes detected.  Closing and unlinking temporary extensions.conf");
            close($efh);
            unlink($efile);
        } else {
            &debug(1, "Changes detected.  Installing extensions.conf");
            rename($efile, $self->{'ASTETCDIR'}.'/extensions.conf');
        }

        $self->_cleanStaleContexts('extensions', \%oldSerials);

        last CONFIG;
    }; # End of /extensions/ case

    /voicemail/ && do {
        my %oldSerials;
        debug(2, "Formatting voicemail.conf");

        # If we aren't flushing the old config get
        # a list of serials for existing contexts
        if (!$self->{'FLUSH'}) {
            %oldSerials = $self->_getOldSerials('voicemail');
        }

        my ($vfh, $vfile) = $self->_getOutputHandle('file', 'voicemail');

        # Process each context and if new write updated data
        CONTEXT: foreach my $context
            (keys(%{$self->{'CONFIGDATA'}{'voicemail'}})) {
            my ($fh, $file);
            &debug(3, "Processing context \"$context\"");


            my $cdata = \%{$self->{'CONFIGDATA'}{'voicemail'}{$context}};
            if (!defined($cdata->{'serial'})) {
                &warn('Serial number not defined for '.$context.': using 0');
                $cdata->{'serial'} = 0;
            }

            # Check to see if we should update the context
            if (defined($oldSerials{$context})) {
                # Remove this context from the list of
                # pre-existing contexts
                if (!$self->{'FLUSH'}) {
                    if ($oldSerials{$context} < $cdata->{'serial'}) {
                        debug(1, "LDAP serial for $context".
                        " ($cdata->{'serial'}) \nis newer than file".
                        " ($oldSerials{$context})");
                    } else {
                        debug(3, "LDAP serial for $context".
                        " ($cdata->{'serial'}) \n is not newer than file".
                        " ($oldSerials{$context}).  Context will not".
                        " be updated.");
                        delete $oldSerials{$context};
                        next CONTEXT;
                    }
                }
            } else {
                &debug(2, 'No old serial number found for '.$context.
                    ' Assuming new context.');
            }
            delete $oldSerials{$context};

            ($fh, $file) =
                $self->_getOutputHandle('file', 'voicemail', $context);

            &debug(2, "Printing context data for $context");
            # Include the context into the master voicemail.conf
            CORE::print $vfh '#include '.$self->{'ASTETCDIR'}.
                "/voicemail.d/voicemail.${context}.conf\n";

            CORE::print $fh "[${context}] ; $cdata->{'serial'}\n";

            &debug(3, 'Printing bare lines');

            foreach my $bareline (@{$cdata->{'barelines'}}) {
                &debug(5, $context.': '.$bareline);
                print $fh "$bareline\n";
            }

            debug(2, "Printing users for $context");
            foreach my $mailbox (keys %{$cdata->{'mailboxes'}}) {
                my $mdata = \%{$cdata->{'mailboxes'}{$mailbox}};

                debug(4, "Printing line for ${mailbox}\@$context");
                print $fh $mailbox.' => '.$mdata->{'pin'}.',';
                print $fh $mdata->{'name'}.',';
                print $fh $mdata->{'email'}.',';
                print $fh $mdata->{'pager'}.',';
                my $looping = 0;
                foreach my $option (@{$mdata->{'options'}}) {
                    if ($looping) { print $fh "|"; } else { $looping = 1; }
                    print $fh "$option";
                }
                print $fh "\n";
            }
            print $fh "\n";

            debug(1, "Wrote all lines for $context, closing and renaming file");
            close($fh);
            rename($file, $self->{'ASTETCDIR'}.'/voicemail.d/voicemail.'.
                $context.'.conf');
            # We've updated a context.  Flag for reload.
            $self->{'NEEDRELOAD'}++;
        } # End of CONTEXT block
        if (!$self->{'NEEDRELOAD'}) {
            &debug(2, 'No changes detected.  Closing and unlinking temporary '.
                'voicemail.conf');
            close($vfh);
            unlink($vfile);
        } else {
            &debug(1, "Changes detected.  Installing voicemail.conf");
            rename($vfile, $self->{'ASTETCDIR'}.'/voicemail.conf');
        }
        $self->_cleanStaleContexts('voicemail', \%oldSerials);

        last CONFIG;
    }; # End of voicemail.conf

    /musiconhold/ && do {
        my $oldSerial;
        my ($mfh, $mfile);
        debug(2, "Formatting musiconhold.conf");

        # Read in old extensions.conf and save serial data
        if (open(OLDEXTENSIONS, '<',
            $self->{'ASTETCDIR'}.'/musiconhold.conf')) {
            while (<OLDEXTENSIONS>) {
                if ($_ =~ /^\[(.*)\]\s+;\s+(\d+)$/) {
                    debug(3, "Found serial number $2 for context $1");
                    $oldSerial = $2;
                } elsif ($_ =~ /^\[(.*)\].*$/) {
                    &warn("Found context with no serial number.  Assuming 0");
                    $oldSerial = 0;
                }
            }
            close(OLDEXTENSIONS);
        } else {
            &warn('Unable to read old musiconhold.conf');
        }

        if (!defined($oldSerial) || !$oldSerial) {
            &debug(1, "Unable to locate old serial number in musiconhold.conf.".
                "  Assuming 0");
            $oldSerial = 0;
        }

        if ($oldSerial >= $self->{'CONFIGDATA'}{'musiconhold'}{'serial'}) {
            &debug(2, "Serial number not newer for musiconhold.  Skipping");
            next CONFIG;
        }

        ($mfh, $mfile) = $self->_getOutputHandle('file', 'musiconhold');
        CORE::print $mfh "[musiconhold] ; ".
            $self->{'CONFIGDATA'}{'musiconhold'}{'serial'}."\n";
        # Process each context and if new write updated data
        CONTEXT: foreach my $context
            (keys(%{$self->{'CONFIGDATA'}{'musiconhold'}{'contexts'}})) {

            &debug(2, "Printing context data for $context");
            # Include the context into the master voicemail.conf
            CORE::print $mfh "$context => ".
                $self->{'CONFIGDATA'}{'musiconhold'}{'contexts'}{$context}."\n";

        }
        debug(1, "Wrote all lines for musiconhold.conf.".
            "  Closing and renaming file");
        close($mfh);
        rename($mfile, $self->{'ASTETCDIR'}.'/musiconhold.conf');
        # We've updated a context.  Flag for reload.
        $self->{'NEEDRELOAD'}++;

        last CONFIG;
    };
    } # End of CONFIG switch
}

=item $AstLDAP->generate($config, $mode)

 Compatibility wrapper for Asterisk::LDAP 0.5.x

 Calls readConfig and format

 Example: $AstLDAP->generate($config, $mode);
=cut
sub generate
{
    my $self = shift(@_);
    my $mode = shift(@_);

    # Default to file-based formatting unless overridden by argument
    if (!defined($mode) || !$mode) {
        $mode = "file";
    }

    foreach my $config (@{$self->{'CONFIGS'}}) {
        $self->readConfig($config);
        $self->format($config, $mode);
    }
}

# Take a config as an option and return a hash of old serials by
# indexed by context
sub _getOldSerials
{
    my $self = shift(@_);
    my $config = shift;
    my %oldSerials;
    # Read in old $config.conf and save serial data
    if (open(OLDEXTENSIONS, '<',
        $self->{'ASTETCDIR'}.'/'.$config.'.conf')) {
        while (<OLDEXTENSIONS>) {
            if ($_ =~ /^\[(.*)\]\s*;\s*(\d+)/) {
                debug(3, "Found serial number $2 for context $1");
                $oldSerials{$1} = $2;
            } elsif ($_ =~ /^\[(.*)\].*$/) {
                &warn("Found context with no serial number.  Assuming 0");
                $oldSerials{$1} = 0;
            }
        }
        close(OLDEXTENSIONS);
    } else {
        &warn('Unable to read old '.$config.'.conf');
    }

    # Check for and possibly attempt to create
    # ASTETCDIR/extensions.d to hold per-context files
    if (!stat($self->{'ASTETCDIR'}.'/'.$config.'.d')) {
        debug(1, 'Creating '.$self->{'ASTETCDIR'}.'/'.$config.'.d');
        mkdir($self->{'ASTETCDIR'}.'/'.$config.'.d') or
            &die('Unable to create '.$self->{'ASTETCDIR'}.
                '/'.$config.'.d for per-context configuration files',
                __FILE__, __LINE__);
    }

    foreach my $context
        (keys(%{$self->{'CONFIGDATA'}{$config}})) {
        if (stat($self->{'ASTETCDIR'}.
            "/${config}.d/${config}.${context}.conf")) {
            open(OLDEXTENSIONS, '<', $self->{'ASTETCDIR'}.
                "/${config}.d/${config}.${context}.conf");
            while (<OLDEXTENSIONS>) {
                if ($_ =~ /^\[(.*)\]\s;\s(\d+)$/) {
                    debug(3, "Found serial number $2 for context $1");
                    $oldSerials{$1} = $2;
                }
            }
            close(OLDEXTENSIONS);
        } else {
            debug(1, "No old configuration file found for $context");
        }
    }

    return %oldSerials;
}

sub _getOutputHandle
{
    my $self = shift(@_);
    my $mode = shift;
    my $config = shift;
    my $context = shift;
    my ($fh, $file);
    my $fileformat;

    # If a context was passed in return a backend for a file-per-context
    # If not return a filehandle for the standard file (ex. extensions.conf)
    if (!defined($context) || !$context) {
        $fileformat = $self->{'ASTETCDIR'}.'/'.$config.'-'.$$.'-XXXXXX';
    } else {
        $fileformat =
            $self->{'ASTETCDIR'}.'/'.$config.'.d/'.$context.'-'.$$.'-XXXXXX';
    }

    # Create a temporary filehandle for the context
    MODE: for ($mode) {
    /file/ && do {
        debug(3, 'Opening temporary file for writing.');
        ($fh, $file) = mkstemp($fileformat) or
            &die('Unable to open temporary file for writing!', __FILE__, __LINE__);
        last MODE;
    };
#     # To properly support scalars as a backend need to figure out how
#     # to handle file-per-context schema.  Maybe just concat all data?
#     # But then we'll have to rip out the #include lines...
#     # Really, does anyone actually *want* this feature?
#     /scalar/ && do {
#         debug(3, "Creating memory backed file for writing");
#         open($fh, '>', \$self->{'OUTPUT'}) or
#             &die( "Unable to open scalar file handle for writing!", __FILE__, __LINE__);
#         last MODE;
#     };
    # default case:
    &die("Invalid or unsupported output format requested");
    } # End of MODE block

    CORE::print $fh "; Generated by Asterisk::LDAP $VERSION\n";
    CORE::print $fh ";\n";
    CORE::print $fh "; !!! DO NOT EDIT !!!\n";
    CORE::print $fh ";\n";

    return ($fh, $file);
}

sub _cleanStaleContexts
{
    my $self = shift(@_);
    my $config = shift;
    my $oldSerials = shift;

    # Any key left in the oldSerials hash is a dead (removed) context
    # which should be purged from disk
    foreach my $oldcontext (keys(%{$oldSerials})) {
        my $filename = $self->{'ASTETCDIR'}."/${config}.d/".
            "${config}.${oldcontext}.conf";

        if(stat($filename)) {
            &print("Pruning old context from disk: ${oldcontext}");
            unlink($filename) or
                &warn("Unable to remove stale context $filename");
        }
        # If the stat fails above the context was probably in extensions.conf
        # itself and was thus overwritten automatically.  No further
        # cleaning should be necessary.
    }
}

=item $AstLDAP->updateVmailPIN([$voicemailbox], [$context], [$pin])

Updates a user's voice mailbox pin number

This method relies on having voicemailbox, context, and newpin either passed
in or previously stored in the object.  This information is then stored to the
directory.  See examples directory with this distribution for a script
compatible with "externpass" option for voicemail.conf.

Example 1:  $AstLDAP->updateVmailPIN("101", "customerA", "12345");

Example 2:  $AstLDAP->voicemailbox("101");
            $AstLDAP->context("customerA");
            $AstLDAP->newpin("12345");
            $AstLDAP->updateVmailPIN();

Example 3:  # Start script with arguments
            # --voicemailbox 101 --context customerA --newpin 12345
            $AstLDAP->parseCmdline();
            $AstLDAP->updateVmailPIN();

Returns: 1 on success.

=cut
sub updateVmailPIN
{
    my $self = shift(@_);
    my $voicemailbox = shift;
    my $context = shift;
    my $newpin = shift;

    my $ldap;
    my $mesg;

    # Check to see if we're missing any needed information.
    # If so, try to pull it in from passed options
    if (!$self->{'VOICEMAILBOX'}) { $self->voicemailbox($voicemailbox); }
    if (!$self->{'CONTEXT}'}) { $self->context($context); }
    if (!$self->{'NEWPIN'}) { $self->newpin($newpin); }

    if (!$self->{'VOICEMAILBOX'} || !$self->{'CONTEXT'} || !$self->{'NEWPIN'}) {
        &die("Must specify voicemailbox, context, and new pin number", __FILE__, __LINE__);
    } else {
        $ldap = $self->ldap_connect("root");
        $mesg = $ldap->search( 'base' => $self->{'USERBASEDN'},
            'filter' => "(&(voiceMailbox=$self->{'VOICEMAILBOX'})(context=$self->{CONTEXT}))" );
        $mesg->code && &die($mesg->error, __FILE__, __LINE__);

        if ($mesg->count != 1) { &die("Mailbox not found!", __FILE__, __LINE__); }

        my $entry = $mesg->entry(0);

        debug(1, "Setting pin for ".$entry->dn()."\n");
        $entry->replace(voiceMailboxPin => $self->{'NEWPIN'});
        $mesg = $entry->update($ldap);
        $mesg->code && &die($mesg->error, __FILE__, __LINE__);
        &print("Pin set for mailbox $self->{'VOICEMAILBOX'} successful.");
    }
    $mesg = $ldap->unbind;
    $mesg->code && &die($mesg->error(), __FILE__, __LINE__);
}

=item $AstLDAP->tryreload()

Attempt to reload Asterisk following a configuration change

This method tests to see if reloading Asterisk is permissable, necessary,
possible, and safe.
To check permissibilty, it looks for a non-zero value in $AstLDAP->astreload();
To check necessity, it looks for a non-zero value in $AstLDAP->needreload();
To check possiblity, it looks for a path to the asterisk binary.
To check safety, it looks for a zero value in $AstLDAP->warningsissued();


Example:    # Generate configuration files and move them into place
            $AstLDAP->needreload(true); # Flag need
            $AstLDAP->astreload(true);  # Flag permission
            $AstLDAP->tryreload();      # Should work if there were no warnings

Returns: 1 on success.

=cut
sub tryreload
{
    my $self = shift(@_);

    if ($self->{'NEEDRELOAD'}) {
        if (!$self->{'ASTERISKBIN'}) {
            &warn("Reload of Asterisk required but no Asterisk binary was found.");
            return;
        }
        if (!$self->{'ASTRELOAD'}) {
            &warn("A manual reload of Asterisk necessary to activate changes.");
            return;
        }
        if (!$main::WARNINGSISSUED) {
            &warn ($main::WARNINGSISSUED." warnings have been issued, will not reload Asterisk.");
            return;
        }

        my $ret = system($self->{'ASTERISKBIN'}, "-rx reload");
        if ($ret) {
            &die("Failed to reload asterisk! system() returned $ret, reason $!", __FILE__, __LINE__);
        } else {
            &print("Reload of Asterisk successful.");
            return 1;
        }
    } else {
        debug(1, "Reload requested but no reload necessary.");
    }
}

=item $AstLDAP->ldap_connect(["root"])

Connect to the LDAP server with configured credentials

This method will connect to the LDAP server with options specified
in the object.  If the optional argument "root" is passed in, ldap_connect
will attempt to bind with privileged credentials.  If none are specified, a
warning will be issued and regular credentials (or anonymous) are tried.


Example: $AstLDAP->ldap_connect();

Returns: A Net::LDAP object.

=cut
sub ldap_connect
{
    my $self = shift(@_);
    my $level = shift;
    my $binddn = undef;
    my $bindpw = undef;

    # Try to find and return an established connection
    if ($self->{'LDAP'}) {
        return $self->{'LDAP'};
    }

    if ($level && $level eq "root") {
        if ($self->{'ROOTBINDDN'} && $self->{'ROOTBINDPW'}) {
            debug(1, "Using ROOT credentials for bind");
            $binddn = $self->{'ROOTBINDDN'};
            $bindpw = $self->{'ROOTBINDPW'};
        } else {
            &warn("Root access requested but either ROOTBINDDN or ROOTBINDPW was unspecified.\nContinuing with default credentials");
            $binddn = $self->{'BINDDN'};
            $bindpw = $self->{'BINDPW'};
        }
    } else {
        $binddn = $self->{'BINDDN'};
        $bindpw = $self->{'BINDPW'};
    }

    # Default connection type is LDAPv3
    my $ldap = "";
    my $i = 0;
    my $mesg;

    if (!$self->{'USERBASEDN'}) {
        debug(2, "No User Base DN specified, assuming ou=People,$self->{'BASEDN'}");
        $self->{'USERBASEDN'} = "ou=People,".$self->{'BASEDN'};
    }
    if (!$self->{'GROUPBASEDN'}) {
        debug(2, "No User Base DN specified, assuming ou=People,$self->{'BASEDN'}");
        $self->{'GROUPBASEDN'} = "ou=Group,".$self->{'BASEDN'};
    }
    while (!$ldap) {
        if ($self->{'LDAPS'}) {
            if (!($self->{'LDAPHOSTS'}[$i] =~ /^ldaps:/)) {
                $self->{'LDAPHOSTS'}[$i] = "ldaps://".$self->{'LDAPHOSTS'}[$i];
            }
        }
        debug(1, "Attempting LDAP connection to ".$self->{'LDAPHOSTS'}[$i]);
        $ldap = Net::LDAP->new($self->{'LDAPHOSTS'}[$i]) or
            &warn("Connection to ".$self->{'LDAPHOSTS'}[$i]." failed!");
        if ($ldap) { last; } else { $i++; }
        if (!$self->{'LDAPHOSTS'}[$i]) {
            &die("Unable to contact any LDAP hosts!", __FILE__, __LINE__);
        }
    }

    # If TLS is requested, attempt to use it
    if ($self->{'STARTTLS'}) {
        debug(1, "Attempting to enable TLS");
        if ($self->{'CACERTDIR'}) {
            debug(2, "Using CA certificates in ".
                $self->{'CACERTDIR'}.", verifying ".$self->{'SSLVERIFY'});
            $ldap->start_tls(
                'capath' => $self->{'CACERTDIR'},
                'verify' => $self->{'SSLVERIFY'}
            );
        } elsif ($self->{'CACERTFILE'}) {
            debug(2, "Using CA certificate ".
                $self->{'CACERTFILE'}.", verifying ".$self->{'SSLVERIFY'});
            $ldap->start_tls(
                'cafile' => $self->{'CACERTFILE'},
                'verify' => $self->{'SSLVERIFY'}
            );
        } else {
            debug(2, "Not using any CA certificates and not verifying the server's certificate.");
            $ldap->start_tls('verify' => 'none');
        }
    }

    if ($binddn) {
        if (!$bindpw) {
            &warn("Authenticated bind requested, but no password given");
        }
        debug(1, "Attempting authenticated bind as $binddn");
        $mesg = $ldap->bind($binddn, password => $bindpw);
        $mesg->code && &die($mesg->error(), __FILE__, __LINE__);
    } else {
        # Anonymous bind
        debug(1, "Attempting anonymous bind");
        $mesg = $ldap->bind;
        $mesg->code && &die($mesg->error(), __FILE__, __LINE__);
    }
    if ($mesg->code) { &die($mesg->error, __FILE__, __LINE__); }

    $self->{'LDAP'} = $ldap;
    return $self->{'LDAP'};
}

=item $AstLDAP->make_backups($filename, $numcopies)

Rotate $AstLDAP->savedcopies() number of backup copies

This method takes as arguments the file to be rotated and a number of copies
to keep.  For example, if "extensions.conf" is passed in, then that file
will be renamed extensions.conf.1; extensions.conf.1 will be renamed to
extensions.conf.2, etc. up to $numcopies.  The last file will be unlinked.
This permits the use of $AstLDAP->undo();


Example: $AstLDAP->make_backups("/etc/asterisk/extensions.conf", 3);

Returns: 1 on success, &die on failure

=cut
sub make_backups
{
    my $self = shift(@_);
    my $filename = shift;
    my $copynum = shift;

    if (!defined($copynum) || $copynum < 1) {
        debug(1, "Backup copy functionality disabled.");
        return 0;
    }

    if (!defined($filename) || !stat($filename)) {
        &die("Backup of file $filename requested, but no file name available.",
            __FILE__, __LINE__);
    } else {
        debug(2, "Backing up \"$filename\"");
    }


    if (stat("${filename}.${copynum}")) {
        debug(2, "Unlinking oldest backup copy of $filename.");
        unlink("${filename}.${copynum}") or
        &die ("Could not unlink ${filename}.${copynum}",__FILE__, __LINE__);
    } else {
        debug(3, "No oldest backup of $filename found.  Removal skipped.");
    }
    $copynum--;
    while ($copynum > 0) {
        my $next = $copynum + 1;
        if (stat("${filename}.${copynum}")) {
            debug(3, "Moving ${filename}.${copynum} to .${next}");
            rename $filename.".".$copynum, $filename.".".$next or
                &die("Unable to rename ${filename}.${copynum}!",
                    __FILE__, __LINE__);
        } else {
            debug(4, "No ${filename}.${copynum} found to rename.");
        }
        $copynum--;
    }
    if (stat("${filename}")) {
        debug(3, "Moving ${filename} to .1");
        rename $filename, $filename.".1" or
            &die("Unable to rename ${filename}!", __FILE__, __LINE__);
    }
}

=item $AstLDAP->undo($filename)

Revert to a backup copy of $filename

This method takes as the sole argument the file to be restored.  For
example, if "/etc/asterisk/extensiosn.conf" is passed in, then that file
will be renamed from <filename>.1 to <filename>

Example: $AstLDAP->undo("/etc/asterisk/extensions.conf");

Returns: 1 on success, &die on failure

=cut
# FIXME Find some way to handle extensions.d en-masse
sub undo
{
    my $self = shift(@_);
    my $filename = shift;;

    &print("Rolling back last update");
    if (stat("${filename}.1")) {
        debug(1, "Moving ${filename}.1 to ${filename}");
        rename $filename.".1", $filename or
            &die("Unable to restore backup file!", __FILE__, __LINE__);
        $self->{'NEEDRELOAD'}++;
        &print("Rollback complete.");
        if (!$self->{'ASTRELOAD'}) {
            &warn("Rollback performed, but no reload requested.");
            &warn("Manually reload Asterisk to activate rollback.");
        }
    } else {
        &die("Backup file not found for ${filename}", __FILE__, __LINE__);
    }
}

sub ldapversion
{
    my $self = shift(@_);
    my $arg = shift;
    my $ldapversion = "";
    if ($arg eq "version") {
        $ldapversion = shift;
        $self->{'CMDLINELDAPVERSION'}++;
    } else {
        $ldapversion = $arg;
    }
    if ($ldapversion) {
        debug(2, "Setting LDAP protocol version to $ldapversion");
        $self->{'LDAPVERSION'} = $ldapversion;
    }
    return $self->{'LDAPVERSION'};
}

sub ldapport
{
    my $self = shift(@_);
    my $arg = shift;
    my $port = "";
    if ($arg eq "port") {
        $port = shift;
        $self->{'CMDLINELDAPPORT'}++;
    } else {
        $port = $arg;
    }
    if ($port) {
        debug(2, "Setting LDAP port to $port");
        $self->{'LDAPPORT'} = $port;
    }
    return $self->{'LDAPPORT'};
}

sub binddn
{
    my $self = shift(@_);
    my $arg = shift;
    my $dn = "";
    if ($arg eq "binddn") {
        $dn = shift;
        $self->{'CMDLINEBINDDN'}++;
    } else {
        $dn = $arg;
    }
    if ($dn) {
        debug(2, "Setting LDAP Bind DN to $dn");
        $self->{'BINDDN'} = $dn;
    }
    return $self->{'BINDDN'};
}

sub bindpw
{
    my $self = shift(@_);
    my $arg = shift;
    my $bindpw = "";
    if ($arg eq "bindpw") {
        $bindpw = shift;
        $self->{'CMDLINEBINDPW'}++;
    } else {
        $bindpw = $arg;
    }
    if ($bindpw) {
        debug(2, "Configuring LDAP Bind password");
        $self->{'BINDPW'} = $bindpw;
    }
    return $self->{'BINDPW'};
}

sub basedn
{
    my $self = shift(@_);
    my $arg = shift;
    my $basedn = "";
    if ($arg eq "basedn") {
        $basedn = shift;
        $self->{'CMDLINEBASEDN'}++;
    } else {
        $basedn = $arg;
    }
    if ($basedn) {
        debug(2, "Setting LDAP Base DN to $basedn");
        $self->{'BASEDN'} = $basedn;
    }
    return $self->{'BASEDN'};
}

sub scope
{
    my $self = shift(@_);
    my $arg = shift;
    my $scope = "";
    if ($arg eq "scope") {
        $scope = shift;
        $self->{'CMDLINESCOPE'}++;
    } else {
        $scope = $arg;
    }
    if ($scope) {
        debug(2, "Setting default LDAP search scope to $scope");
        $self->{'SCOPE'} = $scope;
    }
    return $self->{'SCOPE'};
}

sub ldaps
{
    my $self = shift(@_);
    my $arg = shift;
    my $ldaps = "";
    if ($arg eq "ldaps") {
        $ldaps = shift;
        $self->{'CMDLINELDAPSSL'}++;
    } else {
        $ldaps = $arg;
    }
    if ($ldaps) {
        debug(2, "Will attempt LDAP/SSL connection.");
        $self->{'LDAPS'}++;
    }
    return $self->{'LDAPS'};
}

sub starttls
{
    my $self = shift(@_);
    my $arg = shift;
    my $starttls = "";
    if ($arg eq "tls") {
        $starttls = shift;
        $self->{'CMDLINELDAPTLS'}++;
    } else {
        $starttls = $arg;
    }
    if ($starttls) {
        debug(2, "Will request TLS on LDAP connection.");
        $self->{'STARTTLS'}++;
    }
    return $self->{'STARTTLS'};
}

sub cacertdir
{
    my $self = shift(@_);
    my $arg = shift;
    my $cacertdir = "";
    if ($arg eq "cacertdir") {
        $cacertdir = shift;
        $self->{'CMDLINECACERTDIR'}++;
    } else {
        $cacertdir = $arg;
    }
    if ($cacertdir) {
        debug(2, "Setting CA Certificate directory to $cacertdir");
        $self->{'CACERTDIR'} = $cacertdir;
    }
    return $self->{'CACERTDIR'};
}

sub cacertfile
{
    my $self = shift(@_);
    my $arg = shift;
    my $cacertfile = "";
    if ($arg eq "cacertfile") {
        $cacertfile = shift;
        $self->{'CMDLINECACERTFILE'}++;
    } else {
        $cacertfile = $arg;
    }
    if ($cacertfile) {
        debug(2, "Setting CA Certificate file to $cacertfile");
        $self->{'CACERTFILE'} = $cacertfile;
    }
    return $self->{'CACERTFILE'};
}

sub sslverify
{
    my $self = shift(@_);
    my $arg = shift;
    my $sslverify = "";
    if ($arg eq "sslverify") {
        $sslverify = shift;
        $self->{'CMDLINESSLVERIFY'}++;
    } else {
        $sslverify = $arg;
    }
    if ($sslverify) {
        debug(2, "Enabling SSL Certification Authority validity check.");
        $self->{'SSLVERIFY'}++;
    }
    return $self->{'SSLVERIFY'};
}

sub astetcdir
{
    my $self = shift(@_);
    my $arg = shift;
    my $astetcdir = "";
    if ($arg eq "astetcdir") {
        $astetcdir = shift;
        $self->{'CMDLINEASTETCDIR'}++;
    } else {
        $astetcdir = $arg;
    }
    if ($astetcdir) {
        debug(2, "Setting Asterisk's etc directory to $astetcdir");
        if (stat($astetcdir)) {
            $self->{'ASTETCDIR'} = $astetcdir;
        } else {
            &die("Unable to stat() $astetcdir", __FILE__, __LINE__);
        }
    }
    return $self->{'ASTETCDIR'};
}

sub verbose
{
    my $self = shift(@_);
    if (@_) {
        debug($main::VERBOSE, "Increasing verbosity to ".++$main::VERBOSE);
    }
    return $main::VERBOSE;
}

sub savedcopies
{
    my $self = shift(@_);
    my $arg = shift;
    my $saved = "";
    if ($arg eq "saved") {
        $saved = shift;
    } else {
        $saved = $arg;
    }
    if ($saved) {
        debug(2, "Setting the number of saved copies to $saved");
        $self->{'SAVEDCOPIES'} = $saved;
    }
    return $self->{'SAVEDCOPIES'};
}

sub configs
{
    my $self = shift(@_);
    my $arg = shift;
    my @configs = shift;

    if (defined($arg) && $arg ne "config") {
        # We aren't being called from Getopt. Put the context back on the stack
        $configs[$#configs] = $arg;
    }

    if (@configs) {
        foreach my $config (@configs) {
            if (defined($config) && $config) {
                debug(1, "Adding $config for generation.");
                if (!$self->{'CONFIGOVERRIDE'}) {
                    debug(2, "Configuration override, emptying default list");
                    @{$self->{'CONFIGS'}} = ();
                    $self->{'CONFIGOVERRIDE'}++;
                }
                push(@{$self->{'CONFIGS'}}, $config);
            }
        }
    }
    return @{$self->{'CONFIGS'}};
}

sub astreload
{
    my $self = shift(@_);
    if (@_) {
        debug(1, "Flagging asterisk to reload after update.");
        $self->{'ASTRELOAD'}++;
    }
    return $self->{'ASTRELOAD'};
}

sub asteriskbin
{
    my $self = shift(@_);
    my $arg = shift;
    my $asteriskbin = "";
    if ($arg eq "asteriskbin") {
        $asteriskbin = shift;
    } else {
        $asteriskbin = $arg;
    }
    if ($asteriskbin) {
        debug(2, "Setting path to asterisk binary to $asteriskbin");
        $self->{'ASTERISKBIN'} = $asteriskbin;
    }
    return $self->{'ASTERISKBIN'};
}

sub ldaphosts
{
    my $self = shift(@_);
    my $arg = shift;
    if (!($arg eq "host")) {
        # This will execute if NOT called via Getopt
        push (@_, $arg);
    } else {
        # We must be setting hosts manually.  Flag to ignore ldap.conf hosts
        $self->{'CMDLINELDAPHOSTS'}++;

        # Test if this is the first addition to LDAPHOSTS via the command line.
        # If it is, clear the existing list since we are manually overriding it.
        if ($self->{'CMDLINELDAPHOSTS'} == 1) {
            @{$self->{'LDAPHOSTS'}} = ();
        }
    }
    if (@_) {
        foreach my $host (@_) {
            debug(2, "Adding @_ to the list of LDAP hosts.");
            push(@{$self->{'LDAPHOSTS'}}, @_);
        }
    }
    my $debugstring = "";
    foreach my $host (@{$self->{'LDAPHOSTS'}}) {
        $debugstring .= " $host";
    }
    debug(3, "Have LDAP hosts:$debugstring");
    return $self->{'LDAPHOSTS'};
}

sub needundo
{
    my $self = shift(@_);
    if (@_) {
        debug(2, "Flagging an UNDO procedure.");
        $self->{'UNDO'}++;
    }
    return $self->{'UNDO'};
}

sub quiet
{
    my $self = shift(@_);
    if (@_) {
        $self->{'QUIET'}++;
    }
    return $self->{'QUIET'};
}

sub context
{
    my $self = shift(@_);
    my $arg = shift;
    my $context = "";
    if ($arg eq "context") {
        $context = shift;
        $self->{'CMDLINECONTEXT'}++;
    } else {
        $context = $arg;
    }
    if ($context) {
        debug(2, "Using context $context");
        $self->{'CONTEXT'} = $context;
    }
    return $self->{'CONTEXT'};
}

sub voicemailbox
{
    my $self = shift(@_);
    my $arg = shift;
    my $voicemailbox = "";
    if ($arg eq "voicemailbox") {
        $voicemailbox = shift;
    } else {
        $voicemailbox = $arg;
    }
    if ($voicemailbox) {
        debug(2, "Using voicemailbox $voicemailbox");
        $self->{'VOICEMAILBOX'} = $voicemailbox;
    }
    return $self->{'VOICEMAILBOX'};
}

sub newpin
{
    my $self = shift(@_);
    my $arg = shift;
    my $newpin = "";
    if ($arg eq "newpin") {
        $newpin = shift;
    } else {
        $newpin = $arg;
    }
    if ($newpin) {
        debug(2, "Using new pin $newpin");
        $self->{'NEWPIN'} = $newpin;
    }
    return $self->{'NEWPIN'};
}

sub userbasedn
{
    my $self = shift(@_);
    my $arg = shift;
    my $userbasedn = "";
    if ($arg eq "userbasedn") {
        $userbasedn = shift;
    } else {
        $userbasedn = $arg;
    }
    if ($userbasedn) {
        debug(2, "Setting people base DN to $userbasedn");
        $self->{'USERBASEDN'} = $userbasedn;
    }
    return $self->{'USERBASEDN'};
}

sub groupbasedn
{
    my $self = shift(@_);
    my $arg = shift;
    my $groupbasedn = "";
    if ($arg eq "groupbasedn") {
        $groupbasedn = shift;
    } else {
        $groupbasedn = $arg;
    }
    if ($groupbasedn) {
        debug(2, "Setting people base DN to $groupbasedn");
        $self->{'GROUPBASEDN'} = $groupbasedn;
    }
    return $self->{'GROUPBASEDN'};
}

sub rootbinddn
{
    my $self = shift(@_);
    my $arg = shift;
    my $rootbinddn = "";
    if ($arg eq "rootbinddn") {
        $rootbinddn = shift;
    } else {
        $rootbinddn = $arg;
    }
    if ($rootbinddn) {
        debug(2, "Setting ROOT bind DN to $rootbinddn");
        $self->{'ROOTBINDDN'} = $rootbinddn;
    }
    return $self->{'ROOT'};
}

sub rootbindpw
{
    my $self = shift(@_);
    my $arg = shift;
    my $rootbindpw = "";
    if ($arg eq "rootbindpw") {
        $rootbindpw = shift;
    } else {
        $rootbindpw = $arg;
    }
    if ($rootbindpw) {
        debug(2, "Setting ROOT bind password");
        $self->{'ROOTBINDPW'} = $rootbindpw;
    }
    return $self->{'ROOTBINDPW'};
}

sub ldapconf
{
    my $self = shift(@_);
    my $arg = shift;
    my $ldapconf = "";
    if ($arg eq "ldapconf") {
        $ldapconf = shift;
    } else {
        $ldapconf = $arg;
    }
    if ($ldapconf) {
        debug(2, "Setting ldap.conf to $ldapconf");
        $self->{'LDAPCONF'} = $ldapconf;
    }
    return $self->{'LDAPCONF'};
}

sub asteriskconf
{
    my $self = shift(@_);
    my $arg = shift;
    my $asteriskconf = "";
    if ($arg eq "asteriskconf") {
        $asteriskconf = shift;
    } else {
        $asteriskconf = $arg;
    }
    if ($asteriskconf) {
        debug(2, "Setting asterisk.conf to $asteriskconf");
        $self->{'ASTERISKCONF'} = $asteriskconf;
    }
    return $self->{'ASTERISKCONF'};
}

sub precompile
{
    my $self = shift;
    my $precompile = shift;
    if ($precompile) {
        debug(2, "Precompiling asterisk extensions.conf");
        $self->{'PRECOMPILE'}++;
    }
    return $self->{'PRECOMPILE'};
}

sub needreload
{
    my $self = shift;
    my $needreload = shift;
    if ($needreload) {
        debug(1, "Flagging Asterisk for a reload");
        $self->{'NEEDRELOAD'}++;
    }
    return $self->{'NEEDRELOAD'};
}

sub flush
{
    my $self = shift;
    my $flush = shift;
    if ($flush) {
        debug(2, "Forcing refresh of all contexts");
        $self->{'FLUSH'}++;
    }
    return $self->{'FLUSH'};
}

1;
